"""ArcGIS script that implements the calculated fixed radius capture zone tool

Tested with:
 * ArcGIS 9.3
 * ArcGIS 10.0

Setup as an ArcGIS tool with the following parameters:

Ord: Parameter  : Units : Display Name                : Data Type (options)
---:------------:-------:-----------------------------:------------------------------------
 1 : CZWellPt   :       : Input pump well point       : Feature Layer
 2 : CZPolyFeat :       : Output capture zone polygon : Feature Dataset (Direction: Output)
 3 : Q_LPS      : L/s   : Pump rate                   : Double
 4 : t_years    : years : Travel time                 : Double
 5 : n          : -     : Effective porosity          : Double
 6 : b          : m     : Saturated aquifer thickness : Double

Also add an output parameter:
 Order: 7 (position 6)
 Display name: Output feature
 Data Type: Feature Class
 Type: Derived
 Direction: Output
"""

__author__ = 'Mike Toews'
__email__ = 'M.Toews@gns.cri.nz'
__license__ = 'Proprietary'
__copyright__ = 'Copyright (c) 2013, ' +\
                'Institute of Geological and Nuclear Sciences Limited ' +\
                '(GNS Science)'

import sys
from ggw.ArcGIS import ArcGISProcessor
from ggw.CZ import CalculatedFixedRadius


class CZCalculatedFixedRadius(ArcGISProcessor, CalculatedFixedRadius):
    """Calculated fixed radius capture zone tool for ArcGIS"""
    InputFeatureShapeType = 'Point'
    OutputFeatureIndex = 7
    OutputSchema = [
        ('Q_LPS', 'Q_LPS', 'DOUBLE'),
        ('t_years', 't_years', 'DOUBLE'),
        ('t_days', 't_days', 'DOUBLE'),
        ('n', 'n', 'DOUBLE'),
        ('b', 'b_m', 'DOUBLE'),
        ('radius', 'radius', 'DOUBLE'),
    ]

    def __init__(self, argv):
        """Initialize toolbox with arguments:
            argv[0] : script (ignored)
            argv[1] : CZWellPt
            arvg[2] : CZPolyFeat (output)
            arvg[3] : Q_LPS
            arvg[4] : t_years
            arvg[5] : n
            arvg[6] : b
            argv[7] : Output feature (derived, ignored)
        """
        self.welcome()
        if len(argv) < 7:
            raise Exception('Wrong number of arguments (%d): %s' %
                            (len(argv), argv))
        argv = argv[:]  # copy, so original parameter is not modified
        arg0 = argv.pop(0)  # script (ignored)
        self.InputFeature = argv.pop(0)
        self.sendmsg('    Input pump well point: ' + self.InputFeature)
        self.CZPolyFeat = argv.pop(0)
        self.sendmsg('    Output Capture Zone Polygon: ' + self.CZPolyFeat)
        self.Q_LPS = float(argv.pop(0))
        self.sendmsg('   Q_LPS: %12.3f L/s' % self.Q_LPS)
        self.t_years = float(argv.pop(0))
        self.sendmsg('       t: %12.5g years' % self.t_years)
        self.n = float(argv.pop(0))
        self.sendmsg('       n: %12.3f\t(porosity)' % self.n)
        self.b = float(argv.pop(0))
        self.sendmsg('       b: %12.3f m' % self.b)

        # Read input feature to geometry
        self.read_feature()
        self.Well = self.InputGeometry

        # Build capture zone
        self.build_cz()

        # Save result
        self.save_CZPoly()

        self.sendmsg('\nFinished script.\n')

        return


def main(argv):
    """Run script"""
    p = CZCalculatedFixedRadius(argv)

if __name__ == '__main__':
    """Run script within ArcGIS"""
    main(sys.argv)
