"""ArcGIS script that implements the uniform flow equation capture zone tool

Tested with:
 * ArcGIS 9.3
 * ArcGIS 10.0

Setup as an ArcGIS tool with the following parameters:

Ord: Parameter  : Units : Display Name                : Data Type (options)
---:------------:-------:-----------------------------:------------------------------------
 1 : CZLineFeat :       : Input flow path line        : Feature Layer
 2 : CZPolyFeat :       : Output capture zone polygon : Feature Dataset (Direction: Output)
 3 : K          : m/day : Hydraulic conductivity      : Double
 4 : b          : m     : Saturated aquifer thickness : Double
 5 : i          : m/m   : Hydraulic gradient          : Double
 6 : Q_LPS      : L/s   : Pump rate                   : Double
 7 : n          : -     : Effective porosity          : Double
 8 : time_limit : years : Time limit                  : Double (Type: Optional)

Also add an output parameter:
 Order: 9 (position 8)
 Display name: Output feature
 Data Type: Feature Class
 Type: Derived
 Direction: Output
"""

__author__ = 'Mike Toews'
__email__ = 'M.Toews@gns.cri.nz'
__license__ = 'Proprietary'
__copyright__ = 'Copyright (c) 2013, ' +\
                'Institute of Geological and Nuclear Sciences Limited ' +\
                '(GNS Science)'

import sys
from ggw.ArcGIS import ArcGISProcessor
from ggw.CZ import UniformFlow


class CZUniformFlow(ArcGISProcessor, UniformFlow):
    """Uniform flow equation capture zone tool for ArcGIS"""
    InputFeatureShapeType = 'Polyline'
    OutputFeatureIndex = 8
    OutputSchema = [
        ('K', 'K_mday', 'DOUBLE'),
        ('b', 'b_m', 'DOUBLE'),
        ('i', 'i', 'DOUBLE'),
        ('Q_LPS', 'Q_LPS', 'DOUBLE'),
        ('n', 'n', 'DOUBLE'),
        ('v', 'v', 'DOUBLE'),
        ('T', 'T_m2day', 'DOUBLE'),
        ('Q', 'Q_m3day', 'DOUBLE'),
        ('x0', 'x0', 'DOUBLE'),
        ('ymax', 'ymax', 'DOUBLE'),
        ('ylast', 'ylast', 'DOUBLE'),
        ('t_years', 't_years', 'DOUBLE'),
        ('t_days', 't_days', 'DOUBLE'),
    ]

    def __init__(self, argv):
        """Initialize toolbox with arguments:
            argv[0] : script (ignored)
            argv[1] : CZLineFeat
            arvg[2] : CZPolyFeat (output)
            arvg[3] : K
            arvg[4] : b
            arvg[5] : i
            arvg[6] : Q_LPS
            arvg[7] : n
            arvg[8] : time_limit (optional)
            argv[9] : Output feature (derived, ignored)
        """
        self.welcome()
        if len(argv) < 8:
            raise Exception('Wrong number of arguments (%d): %s' %
                            (len(argv), argv))
        argv = argv[:]  # copy, so original parameter is not modified
        arg0 = argv.pop(0)  # script (ignored)
        self.InputFeature = argv.pop(0)
        self.sendmsg('    Input Flow Path Line: ' + self.InputFeature)
        self.CZPolyFeat = argv.pop(0)
        self.sendmsg('    Output Capture Zone Polygon: ' + self.CZPolyFeat)
        self.K = float(argv.pop(0))
        self.sendmsg('       K: %12.5g m/day' % self.K)
        self.b = float(argv.pop(0))
        self.sendmsg('       b: %12.3f m' % self.b)
        self.i = float(argv.pop(0))
        self.sendmsg('       i: %12.3g m/m' % self.i)
        self.Q_LPS = float(argv.pop(0))
        self.sendmsg('   Q_LPS: %12.3f L/s' % self.Q_LPS)
        self.n = float(argv.pop(0))
        self.sendmsg('       n: %12.3f\t(porosity)' % self.n)
        try:
            time_limit = float(argv.pop(0))
        except:  # KeyError or ValueError ("#" is default)
            time_limit = None
        self.sendmsg(' time limit: %s years\t(optional)' % str(time_limit))

        # Read input feature to geometry
        self.read_feature()
        self.FlowLine = self.InputGeometry

        # Build capture zone
        self.build_cz(time_limit)

        # Save result
        self.save_CZPoly()

        self.sendmsg('\nFinished script.\n')

        return


def main(argv):
    """Run script"""
    p = CZUniformFlow(argv)

if __name__ == '__main__':
    """Run script within ArcGIS"""
    main(sys.argv)
