"""Core utilities for ArcGIS implementation of GNS Groundwater Tools
"""

__author__ = 'Mike Toews'
__email__ = 'M.Toews@gns.cri.nz'
__license__ = 'Proprietary'
__copyright__ = 'Copyright (c) 2013, ' +\
                'Institute of Geological and Nuclear Sciences Limited ' +\
                '(GNS Science)'

import os
import sys
from textwrap import dedent

import arcgisscripting
from .core import GeoProcessor, VERBOSITY, Point, LineString

# This implies ArcGIS 9.3, but it appears to be
# compatible with other versions
gp = arcgisscripting.create(9.3)
gp.toolbox = 'management'

from distutils.version import StrictVersion
ArcGISVersion = StrictVersion(gp.getinstallinfo('Desktop')['Version'])


class ArcGISProcessor(GeoProcessor):
    """Abstract class for processing capture zones with ArcGIS"""

    ## Set these for an inherited class
    InputFeatureShapeType = None  # [Polyline, Point]

    def sendmsg(self, msg, level=1):
        """Show a message in the Processing Dialog in ArcToolbox"""
        if level <= VERBOSITY:
            gp.AddMessage(msg)

    def sendwarn(self, msg=None, id=None, level=1):
        """Show a warning in the Processing Dialog in ArcToolbox"""
        if level <= VERBOSITY:
            if msg is not None:
                msg = 'Warning: ' + msg
                gp.AddWarning(msg)
            if id is not None:
                gp.AddIDMessage('Warning', id)

    def senderror(self, msg=None, id=None):
        """Show a error in the Processing Dialog in ArcToolbox. If there
        is an error ID, then raise this"""
        if msg is not None:
            msg = 'Error: ' + msg
            gp.AddError(msg)
        if id is not None:
            gp.AddIDMessage('Error', id)
        raise Exception(msg)

    def read_feature(self):
        """Read ArcGIS feature from self.InputFeature,
        and store it as a Shapely geometry at self.InputGeometry
        """
        self.sendmsg('\nReading Input Feature...')
        self.sendmsg('          Feature : ' + self.InputFeature)
        desc = gp.Describe(self.InputFeature)

        # Try to get workspace from the input feature
        if os.path.isfile(self.InputFeature):
            self.workspace = os.path.dirname(self.InputFeature)
        if not hasattr(self, 'workspace') or self.workspace == '':
            self.workspace = os.path.dirname(desc.CatalogPath)
        self.sendmsg('        Workspace : ' + self.workspace)
        gp.workspace = self.workspace

        # This needs to match expected InputFeatureShapeType
        ShapeType = desc.ShapeType
        self.sendmsg('        ShapeType : ' + str(ShapeType))
        if ShapeType != self.InputFeatureShapeType:
            msg = "ShapeType: Found '%s' but expected '%s'" %\
                  (ShapeType, self.InputFeatureShapeType)
            self.senderror(msg, 913)
            return

        # 3D feature
        self.HasZ = desc.HasZ
        self.sendmsg('             HasZ : ' + str(self.HasZ))

        # Geometry field name, spatial reference
        ShapeFieldName = desc.ShapeFieldName
        self.sendmsg('   ShapeFieldName : ' + str(ShapeFieldName))
        self.SpatialReference = desc.SpatialReference

        # Make sure it is projected in metres
        if self.SpatialReference:
            ProjectionType = self.SpatialReference.Type
            self.sendmsg('   ProjectionType : ' + str(ProjectionType))
            if ProjectionType == 'Projected':
                pass  # perfect, it should be projected
            elif ProjectionType == 'Geographic':
                self.senderror('Input feature must be Projected. '
                               'Found Geographic projection type.')
                return
            else:  # it is probably just blank
                msg = "Spatial reference type is not 'Projected'; "\
                      "found %r" % str(ProjectionType)
                if ArcGISVersion >= '10.0':
                    id = 1022
                else:
                    id = None
                self.sendwarn(msg, id)
            LinearUnitName = self.SpatialReference.LinearUnitName
            self.sendmsg('   LinearUnitName : ' + str(LinearUnitName))
            if LinearUnitName != 'Meter':
                msg = dedent("""\
                    Linear Unit Name is not 'Meter'; found %r.
                    All calculations performed assume map units of meters,
                    let's hope they are actually in meters.""" %
                             str(LinearUnitName))
                self.sendwarn(msg)
        else:  # no SpatialReference
            self.sendwarn(id=248)
        # Release description
        del desc

        # Get geometry feature
        FeatureRows = gp.SearchCursor(self.InputFeature)

        # There should only be one row, one feature, one part
        FeatureRow = FeatureRows.Next()
        if not FeatureRow:
            # No features found in <value>. Possible empty feature class.
            gp.AddIDMessage('Error', 10151, 'input capture zone feature')
            return
        FeatureGeometry = FeatureRow.GetValue(ShapeFieldName)
        assert FeatureGeometry, 'No FeatureGeometry'
        FeatureGeometryPart = FeatureGeometry.GetPart()
        assert FeatureGeometryPart, 'No FeatureGeometryPart'

        # Convert the Esri shape to Shapley geometry
        if ShapeType == 'Polyline':
            Geometry = FeatureGeometryPart.Next()
            assert Geometry, 'No Geometry'
            # Collect all vertex coordinates
            pts = []
            while True:
                pt = Geometry.Next()
                if not pt:
                    break
                if self.HasZ:
                    pts.append((pt.X, pt.Y, pt.Z))
                else:
                    pts.append((pt.X, pt.Y))
            del pt
            assert len(pts) > 1, len(pts)
            self.InputGeometry = LineString(pts)
            Geometry = FeatureGeometryPart.Next()
            if Geometry:
                self.sendwarn('Only the first part of the feature '
                              'geometry is used. Remaining parts ignored.')
            del Geometry
        elif ShapeType == 'Point':
            pt = FeatureGeometryPart
            if self.HasZ:
                self.InputGeometry = Point(pt.X, pt.Y, pt.Z)
            else:
                self.InputGeometry = Point(pt.X, pt.Y)
            del pt
        else:
            raise Exception("not sure how to convert '%s'" % ShapeType)

        # Make sure there are no other geometry info
        # otherwise show warnings that they are ignored
        FeatureRow = FeatureRows.Next()
        if FeatureRow:
            self.sendwarn('Only the first geometry is used. '
                          'Remaining geometries ignored.')
        del FeatureRow

        # Release the feature cursor
        del FeatureRows
        del FeatureGeometryPart

        return

    def save_CZPoly(self):
        """Save CZPoly feature back to ArcGIS -- geometry and attributes

        Capture Zone Feature:
            Geometry: self.CZPolyFeat Shapely Polygon
            Attributes: self.OutputSchema list of tuples:
                [(self_attr_name, feat_attr_name, feat_attr_dtype),
                ...]

        Write feature to self.CZPolyFeat, which could be, e.g., a shapefile
        """

        # Change workspace if it is different
        test_dir = os.path.dirname(self.CZPolyFeat)
        if os.path.isdir(test_dir):
            self.workspace = test_dir
            if gp.workspace != test_dir:
                gp.workspace = test_dir

        # see if CZPolyFeat exists
        if gp.Exists(self.CZPolyFeat):
            self.sendmsg('\nUsing existing CZPolyFeat: ' +
                         str(self.CZPolyFeat))
        else:  # Feat does not exist, create it
            self.CZPolyFeat = os.path.basename(self.CZPolyFeat)
            self.sendmsg('\nCreating feature class...')
            self.sendmsg('   Workspace: ' + self.workspace)
            self.sendmsg('  CZPolyFeat: ' + self.CZPolyFeat)

            gp.CreateFeatureclass(self.workspace, self.CZPolyFeat, 'POLYGON',
                                  '', '', '', self.SpatialReference)

            self.sendmsg('Adding fields...')
            for attr in self.OutputSchema:
                assert len(attr) == 3, attr
                gp.AddField(self.CZPolyFeat, attr[1], attr[2])

        self.sendmsg('Fetching insert cursor...')
        CZPolyRows = gp.InsertCursor(self.CZPolyFeat)
        if not CZPolyRows:
            raise Exception('No CZPolyRows')

        # Build ArcObject
        self.sendmsg('Building feature...')
        feat = CZPolyRows.NewRow()
        # Attribute data
        for attr in self.OutputSchema:
            feat.SetValue(attr[1], getattr(self, attr[0]))
        # Geometry
        self.sendmsg('Building geometry...')
        ptObj = gp.CreateObject('Point')
        arrayObj = gp.CreateObject('Array')
        for pt in self.CZpoly.boundary.coords:
            ptObj.x = pt[0]
            ptObj.y = pt[1]
            #if self.HasZ:
            #    ptObj.z = pt[2]
            arrayObj.add(ptObj)
        feat.Shape = arrayObj
        # Add the capture zone
        CZPolyRows.InsertRow(feat)

        # Cleanup
        arrayObj.RemoveAll()
        del CZPolyRows

        # Send feature class back to ArcGIS as a derived output parameter
        try:
            gp.SetParameterAsText(self.OutputFeatureIndex, self.CZPolyFeat)
        except:
            self.sendwarn('Could not set parameter %r value %r' %
                          (self.OutputFeatureIndex, self.CZPolyFeat))

        return


def main(argv):
    """Run script"""
    p = ArcGISCZBearJacobs1965(argv)


if __name__ == '__main__':
    """Start of script within ArcGIS"""
    main(sys.argv)
