"""GNS Groundwater Capture Zone Tools
"""

__author__ = 'Mike Toews'
__email__ = 'M.Toews@gns.cri.nz'
__credits__ = 'Maksym Gusyev, Conny Tschritter, Magali Moreau-Fournier'
__license__ = 'Proprietary'
__copyright__ = 'Copyright (c) 2013, ' +\
                'Institute of Geological and Nuclear Sciences Limited ' +\
                '(GNS Science)'

__all__ = ['UniformFlow', 'CalculatedFixedRadius']

from .core import GeoProcessor, isnumeric, \
    Point, LineString, Polygon, TopologicalError
from textwrap import dedent
from math import pi, atan2, sin, cos, tan, sqrt, log


class CalculatedFixedRadius(GeoProcessor):
    """Delineates a calculated fixed radius capture zone

    This is actually a relatively simple method, where a point
    is buffered by a radius (US EPA 1987):

        radius = sqrt( (Q * t) / (pi * n * b) )

    where:
        Q is the pump rate [L^3/T]
        t is the travel time [T]
        n is the effective porosity [L^3/L^3]
        b is the saturated aquifer thickness or the length of well screen [L]

    Example:
    >>> from ggw.CZ import Point, CalculatedFixedRadius
    >>> pt = Point(2693226, 5988760)
    >>> # initialise with 2 years
    >>> cfr = CalculatedFixedRadius(pt, 10, 2, 0.2, 30)
    >>> cfr.build_cz()
    >>> # Recalculate for 5 years
    >>> cfr.build_cz(5)
    """

    def _getWell(self):
        """Point object for well location"""
        return self._Well

    def _setWell(self, Well):
        if not isinstance(Well, Point):
            raise TypeError('Well input must be a Shapely Point object')
        elif Well.is_empty:  # Make sure the geometry is good
            raise ValueError('Point is empty')
        self._Well = Well

    Well = property(_getWell, _setWell)

    def _getQ_LPS(self):
        """Pump rate [L/s]"""
        return self._Q_LPS

    def _setQ_LPS(self, Q_LPS):
        Q_LPS = float(Q_LPS)
        if not Q_LPS > 0:
            raise ValueError("Pump rate 'Q_LPS' must be greater than zero")
        self._Q_LPS = Q_LPS

    Q_LPS = property(_getQ_LPS, _setQ_LPS)

    def _getQ(self):
        """Pump rate [m/day]"""
        return self._Q_LPS * 86.4

    def _setQ(self, Q):
        self.Q_LPS = float(Q) / 86.4

    Q = property(_getQ, _setQ)

    def _gett_years(self):
        """Travel time [years]"""
        return self._t_years

    def _sett_years(self, t_years):
        t_years = float(t_years)
        if not t_years > 0:
            raise ValueError("Travel time 't_years' must be greater than zero")
        self._t_years = t_years

    t_years = property(_gett_years, _sett_years)

    def _gett_days(self):
        """Travel time [days]"""
        return self._t_years * self.days_per_year

    def _sett_days(self, t_days):
        self.t_years = float(t_days) / self.days_per_year

    t_days = property(_gett_days, _sett_days)

    def _getn(self):
        """Effective porosity [-]"""
        return self._n

    def _setn(self, n):
        n = float(n)
        if not (n > 0 and n < 1):
            raise ValueError("Effective porosity 'n' must be "
                             "between 0 and 1 (exclusive)")
        self._n = n

    n = property(_getn, _setn)

    def _getb(self):
        """Saturated aquifer thickness or the length of well screen [m]"""
        return self._b

    def _setb(self, b):
        b = float(b)
        if not b > 0:
            raise ValueError("Saturated aquifer thickness 'b' must be "
                             "greater than zero")
        self._b = b

    b = property(_getb, _setb)

    @property
    def radius(self):
        """Calculated fixed radius [m]"""
        return sqrt((self.Q * self.t_days) / (pi * self.n * self.b))

    def __init__(self, Well, Q_LPS, t_years, n, b):
        """Initialise required parameters"""
        self.Well = Well
        self.Q_LPS = Q_LPS
        self.t_years = t_years
        self.n = n
        self.b = b
        self.show_parameters()

    def show_parameters(self):
        """Show parameters"""
        self.sendmsg('\nInput parameters...')
        self.sendmsg('Well location : ' + self.Well.wkt)
        self.sendmsg('          Q : %.3f L/s or %.1f m^3/day' %
                     (self.Q_LPS, self.Q))
        self.sendmsg('          t : %.3f years or %.1f days' %
                     (self.t_years, self.t_days))
        self.sendmsg('          n : %.4f\t(porosity)' % self.n)
        self.sendmsg('          b : %.3f m' % self.b)
        self.sendmsg('\nCalculating fixed radius...')
        self.sendmsg('  radius : %12.3f m' % self.radius)

    def build_cz(self, t_years=None):
        """Build capture zone polygon

        t_years can be provided to build capture zones for different
        travel times.
        """
        if t_years is not None:
            self.t_years = t_years
            self.show_parameters()

        self.sendmsg('\nBuilding capture zone...\n')

        # Combine capture zone points
        self.sendmsg('Buffering capture zone around well...')
        self.CZpoly = self.Well.buffer(self.radius, self.curve_resolution)

        self.sendmsg('   Capture zone WKT:\n' + self.CZpoly.wkt)
        self.sendmsg('   Number of vertices: %i' %
                     (len(self.CZpoly.boundary.coords) - 1,))
        self.sendmsg('   Capture zone area: %.1f m^2 or %.3f hectares' %
                     (self.CZpoly.area, self.CZpoly.area / 10000.0))

        # Check output
        if self.CZpoly.is_empty:
            raise ValueError('Capture zone Polygon is empty')
        if not self.CZpoly.is_valid:
            self.sendwarn('Capture zone Polygon is not valid')
        if not self.CZpoly.is_simple:
            raise ValueError('Capture zone Polygon is not simple')

        # Done!
        return


class UniformFlow(GeoProcessor):
    """Calculate a capture zone using the uniform flow equation method

    This method is defined as (Bear and Jacobs 1965; Todd 1980):

        x = -y / tan(y * 2 * pi * Q0 / Q)

    where:
        Q0 is the ambient uniform flow in the aquifer in the x-direction
        x and y are distances from the pumping well at the origin
        Q is the pump rate

    Example:
    >>> from shapely.wkt import loads as load_wkt
    >>> from ggw.CZ import UniformFlow
    >>> ls = load_wkt('LINESTRING(0 0 0, 1000 0 10)')
    >>> cz = UniformFlow(ls, 20, 30, 0.006, 20, 0.2)
    >>> cz.build_cz()
    >>> cz.build_cz(time_limit=2)
    """

    def _getFlowLine(self):
        """LineString object for ambient groundwater flow to a well

        Note that this object could be modified by 'time_limit', so an
        original copy is found at _OrigFlowLine"""
        return self._FlowLine

    # Backup, which is set on the first setter on FlowLine
    _OrigFlowLine = None

    def _setFlowLine(self, FlowLine):
        if not isinstance(FlowLine, LineString):
            raise TypeError('Capture zone input must be a '
                            'Shapely LineString object')
        elif FlowLine.is_empty:  # Make sure the geometry is good
            raise ValueError('LineString is empty')
        elif not FlowLine.is_valid:
            raise ValueError('LineString is not valid')
        elif not FlowLine.is_simple:
            raise ValueError('LineString is not simple, i.e., it probably '
                             'crosses itself')
        elif FlowLine.is_ring:
            raise ValueError('LineString is a ring (or is closed)')
        if self._OrigFlowLine is None:
            # Make a backup of this object on the first read
            self._OrigFlowLine = FlowLine
        self._FlowLine = FlowLine

    FlowLine = property(_getFlowLine, _setFlowLine)

    def _getK(self):
        """Hydraulic conductivity [m/day]"""
        return self._K

    def _setK(self, K):
        K = float(K)
        if not K > 0:
            raise ValueError("Hydraulic conductivity 'K' must be "
                             "greater than zero")
        self._K = K

    K = property(_getK, _setK)

    def _getb(self):
        """Saturated aquifer thickness [m]"""
        return self._b

    def _setb(self, b):
        b = float(b)
        if not b > 0:
            raise ValueError("Saturated aquifer thickness 'b' must be "
                             "greater than zero")
        self._b = b

    b = property(_getb, _setb)

    def _geti(self):
        """Hydraulic gradient [m/m]"""
        return self._i

    def _seti(self, i):
        i = float(i)
        if not i > 0:
            raise ValueError("Hydraulic gradient 'i' must be "
                             "greater than zero")
        self._i = i

    i = property(_geti, _seti)

    def _getQ_LPS(self):
        """Pump rate [L/s]"""
        return self._Q_LPS

    def _setQ_LPS(self, Q_LPS):
        Q_LPS = float(Q_LPS)
        if not Q_LPS > 0:
            raise ValueError("Pump rate 'Q_LPS' must be greater than zero")
        self._Q_LPS = Q_LPS

    Q_LPS = property(_getQ_LPS, _setQ_LPS)

    def _getQ(self):
        """Pump rate [m/day]"""
        return self._Q_LPS * 86.4

    def _setQ(self, Q):
        self.Q_LPS = float(Q) / 86.4

    Q = property(_getQ, _setQ)

    def _getn(self):
        """Effective porosity [-]"""
        return self._n

    def _setn(self, n):
        n = float(n)
        if not (n > 0 and n < 1):
            raise ValueError("Effective porosity 'n' must be "
                             "between 0 and 1 (exclusive)")
        self._n = n

    n = property(_getn, _setn)

    # Derived parameters

    @property
    def T(self):
        """Transmissivity [m^2/day]

        Calculated using:
            T = K*b
        """
        return self.K * self.b

    @property
    def v(self):
        """Average groundwater velocity [m/day]

        Calculated using Darcy's law:
            v = K*i/n
        """
        return self.K * self.i / self.n

    @property
    def ymax(self):
        """Maximum capture zone half-width [m]

        Calculated using:
            ymax = Q / (2*K*b*i)
        """
        return self.Q / (2.0 * self.K * self.b * self.i)

    @property
    def Q0(self):
        """Ambient uniform flow in the aquifer in x-direction [m^2/day]

        Calculated using:
            Q0 = K*b*i
        """
        return self.K * self.b * self.i

    @property
    def x0(self):
        """Downstream stagnation distance [m]

        Calculated using:
            x0 = -Q/(2*pi*K*b*i)
        """
        return -self.Q / (2.0 * pi * self.K * self.b * self.i)

    @property
    def Well(self):
        """Well location; first point from FlowLine"""
        return Point(self.FlowLine.coords[0])

    @property
    def t_years(self):
        """Travel time along FlowLine [years]"""
        rx = self.FlowLine.length
        return self.tx_days(rx) / self.days_per_year

    @property
    def t_days(self):
        """Travel time along FlowLine [days]"""
        rx = self.FlowLine.length
        return self.tx_days(rx)

    def tx_days(self, rx):
        """Travel time [days] to an upstream distance along flow path, rx

        Calculated using:
            tx = (1/v) * (rx + x0 * ln(1 - rx / x0))
        """
        return (rx + self.x0 * log(1 - rx / self.x0)) / self.v

    def tx_years(self, rx):
        """Travel time [years] to an upstream distance along flow path, rx

        See tx_days for calculation"""
        return self.tx_days(rx) / self.days_per_year

    def rx_days(self, tx):
        """Upstream distance along flow path to a point given tx [days]

        Inverse function of tx_days, by iteration using a secant method
        """

        def secant(func, oldx, x, TOL=1e-6):
            """Modified secant method
            http://www.physics.rutgers.edu/~masud/computing/WPark_recipes_in_python.html#Sec4.2
            """
            oldf, f = func(oldx), func(x)
            if (abs(f) > abs(oldf)):  # swap so that f(x) is closer to 0
                oldx, x = x, oldx
                oldf, f = f, oldf
            count = 0
            while True:
                self.sendmsg('secant(%d): x=%s, f(x)=%s' %
                             (count, x, f), level=2)
                try:
                    dx = f * (x - oldx) / float(f - oldf)
                except ZeroDivisionError:
                    self.sendwarn('No slope, this must be 0', level=2)
                    return 0.0
                if abs(dx) < TOL * (1 + abs(x)):
                    return x - dx
                oldx, x = x, x - dx
                # x MUST be greater than 0.0
                if x <= 0.0:
                    self.sendwarn('Bumping x from %s to 0.0' % x, level=2)
                    x = 0.0
                oldf, f = f, func(x)
                count += 1
        # Calculate shifted root, staring with initial
        # guess first rx using average velocity
        return secant(lambda xx: tx - self.tx_days(xx),
                      0.0, tx * self.v)

    def rx_years(self, tx):
        """Upstream distance along flow path to a point given tx [years]

        See rx_days for method
        """
        return self.rx_days(tx * self.days_per_year)

    ## Uniform Flow Equations
    # dslf :: dividing streamline function

    def dslf(self, y):
        """dslf(y) : find x-distance given capture zone half-width y

        Calculated using:
            x = -y/tan(2*pi*K*b*i*y/Q)
        """
        if not (y < self.ymax):
            raise ValueError("'y' must be less than 'ymax'")
        return -y / tan(2.0 * pi * self.K * self.b * self.i * y / self.Q)

    def dslf_i(self, x):
        """Inverse function of dslf, by iteration

        dslf_i(x) : find capture zone half-width, y, given x-distance
        """

        def secant(func, oldx, x, TOL=1e-6):
            """Modified secant method
            http://www.physics.rutgers.edu/~masud/computing/WPark_recipes_in_python.html#Sec4.2
            """
            oldf, f = func(oldx), func(x)
            if (abs(f) > abs(oldf)):  # swap so that f(x) is closer to 0
                oldx, x = x, oldx
                oldf, f = f, oldf
            count = 0
            while True:
                self.sendmsg('secant(%d): x=%s, f(x)=%s' %
                             (count, x, f), level=2)
                try:
                    dx = f * (x - oldx) / float(f - oldf)
                except ZeroDivisionError:
                    self.sendwarn('No slope, this must be ymax', level=2)
                    return self.ymax
                if abs(dx) < TOL * (1 + abs(x)):
                    return x - dx
                oldx, x = x, x - dx
                # x MUST be in the exclusive interval (0,ymax)
                if x <= 0.0:
                    # 1% of ymax
                    self.sendwarn('Trimming x from %s to %s' %
                                  (x, self.ymax * 0.01), level=2)
                    x = self.ymax * 0.001
                elif x >= self.ymax:
                    # 99.9% of ymax
                    self.sendwarn('Trimming x from %s to %s' %
                                  (x, self.ymax * 0.999), level=2)
                    x = self.ymax * 0.999
                oldf, f = f, func(x)
                count += 1
        # Calculate shifted root, staring with initial
        # guess +/-10% to the perpendicular well location (ymax/2.0)
        return secant(lambda xx: x - self.dslf(xx),
                      self.ymax / 2.2, self.ymax / 1.8)

    def __init__(self, FlowLine, K, b, i, Q_LPS, n):
        """Initialise required parameters"""
        self.FlowLine = FlowLine
        self.K = K
        self.b = b
        self.i = i
        self.Q_LPS = Q_LPS
        self.n = n
        self.show_parameters()

    def show_parameters(self):
        """Show parameters"""
        self.sendmsg('\nInput parameters...')
        self.sendmsg('FlowLine : %s' % repr(self.FlowLine))
        self.sendmsg('       K : %.5g m/day' % self.K)
        self.sendmsg('       b : %.3f m' % self.b)
        self.sendmsg('       i : %.3g m/m' % self.i)
        self.sendmsg('       Q : %.3f L/s or %.1f m^3/day' %
                     (self.Q_LPS, self.Q))
        self.sendmsg('       n : %.4f\t(porosity)' % self.n)
        self.sendmsg('\nCalculating derived parameters...')
        self.sendmsg('       T : %.5g m^2/day' % self.T)
        self.sendmsg('       v : %.3f m/day' % self.v)
        self.sendmsg('      Q0 : %.3f m^2/day\t'
                     '(ambient uniform flow)' % self.Q0)
        self.sendmsg('      x0 : %.3f m\t'
                     '(downstream stagnation distance)' % self.x0)
        self.sendmsg('    ymax : %.3f m\t'
                     '(maximum capture zone half-width)' % self.ymax)
        self.sendmsg('   Well location : ' + self.Well.wkt)
        self.sendmsg('Flow Line length : %.3f m' % self.FlowLine.length)
        if self.t_years is None:
            self.sendmsg('     travel time : None' %
                         (self.t_years, self.t_days))
        else:
            self.sendmsg('     travel time : %.3f years or %.1f days' %
                         (self.t_years, self.t_days))

    def build_cz(self, time_limit=None):
        """Build capture zone polygon

        time_limit can be provided to build capture zones for different
        travel times.
        """
        # Restore original FlowLine, in case it was modified by
        # a previous time_limit setting
        self.FlowLine = LineString(self._OrigFlowLine.coords[:])

        self.sendmsg('\nBuilding capture zone...\n')

        ## Internal helper functions
        def accumulate_cz_points(ptx, y, theta, side=None):
            """Add point(s) to CZ point array with distance y and direction
            normal to theta. If side is None, both left and right sides are
            added, otherwise it will only use one or the other."""
            assert side in ('left', 'right', None), side
            # translate and rotate the points
            if side is None or side == 'left':
                self.CZpts.insert(0, (ptx.x + y * cos(theta + pi / 2.0),
                                      ptx.y + y * sin(theta + pi / 2.0)))
            if side is None or side == 'right':
                self.CZpts.append((ptx.x + y * cos(theta - pi / 2.0),
                                   ptx.y + y * sin(theta - pi / 2.0)))
            return

        def dslf_buf(bufpts, side):
            """Adjust a buffer on a side"""
            assert side in ('left', 'right'), side
            count_main_pt = 0
            count_refine = 0
            # start with start of line, at well
            bpA = Point(bufpts[0])
            xA = self.FlowLine.project(bpA)
            # should be at or near well
            assert abs(xA - 0.0) < 0.0001, (xA, bufpts[0])
            yA = self.ymax / 2.0
            lpA = self.FlowLine.interpolate(xA)

            # Go though each point in the buffer
            for vert_i in range(len(bufpts) - 1):
                # Evaluate next buffer point
                bpB = Point(bufpts[vert_i + 1])
                xB = self.FlowLine.project(bpB)
                yB = self.dslf_i(xB)
                lpB = self.FlowLine.interpolate(xB)

                # Direction FlowLine, as calculated
                # from the tangent of the buffer projection
                if side == 'left':
                    theta = atan2(lpB.y - bpB.y, lpB.x - bpB.x) + pi / 2.0
                elif side == 'right':
                    theta = atan2(lpB.y - bpB.y, lpB.x - bpB.x) - pi / 2.0

                # Evaluate the difference in the width
                ydiff = yB - yA
                # Refine more points in-between vertices, if required
                # See how many (if any) dy can fit in there
                nextra = ydiff / dy

                if int(nextra) > 0:
                    # Refine slope
                    dym = ydiff / int(nextra)

                    for yi in range(int(nextra) - 1):
                        # calculate simple x and y
                        yt = dym * (yi + 1) + yA
                        xt = self.dslf(yt)
                        # interpolate the point along the capture zone line
                        ptx = self.FlowLine.interpolate(xt)
                        accumulate_cz_points(ptx, yt, theta, side)
                        count_refine += 1

                # Add the buffer point, adjusting for upstream distance
                accumulate_cz_points(lpB, yB, theta, side)
                count_main_pt += 1

                # Use for next vertex point
                bpA = bpB
                lpA = lpB
                xA = xB
                yA = yB

            self.sendmsg('Added %i main points, %i refined points to %s side' %
                         (count_main_pt, count_refine, side), level=2)
            return

        def build_side(side):
            """Build lines for the 'left' or 'right' sides"""
            assert side in ('left', 'right')
            self.sendmsg('Building %s side buffer with %.3f m...' %
                         (side, yt2))
            try:
                buf = self.FlowLine.parallel_offset(yt2, side, resolution=self.curve_resolution)
            except TopologicalError:
                raise TopologicalError(topomsg % side)
            except ValueError:  # older shapely versions
                buf = LineString()  # empty, raise Exception below
            self.sendmsg(buf.wkt, level=2)
            if buf.type != 'LineString':
                msg = 'Unexpected resulting shape type: ' + str(buf.type)
                raise TopologicalError(msg + '\n' + (topomsg % side))
            elif buf.is_empty:
                raise TopologicalError(topomsg % side)
            buf_coords = list(buf.coords)
            self.sendmsg('Adding %d upstream points on %s side...' %
                         (len(buf_coords), side))
            # Coords from right side need to be flipped
            if side == 'right':
                buf_coords.reverse()
            dslf_buf(buf_coords, side)
            return

        # Optional variable
        if time_limit is not None:
            if not (isnumeric(time_limit) and time_limit > 0):
                raise ValueError("'time_limit' must be greater than zero")
            elif self.t_years <= time_limit:
                self.sendmsg('Calculated travel time for input flow line is '
                             'less than provided time_limit')
                self.sendmsg('Capture line path does not need to be modified '
                             'for time_limit')
            else:  # self.t_years > time_limit:
                self.sendmsg('Modifying ambient flow line path to fit '
                             'time limit...')
                # Calculate fraction of line length needed
                frac = self.rx_years(time_limit) / self.FlowLine.length
                assert frac <= 1.0 and frac > 0.0, frac
                end_c = self.FlowLine.interpolate(frac, True).coords[0]
                for end_i, c in enumerate(self.FlowLine.coords):
                    if c == end_c:
                        end_c = None
                        end_i += 1
                        break
                    if self.FlowLine.project(Point(c), True) >= frac:
                        break
                extracted = self.FlowLine.coords[0:end_i]
                if end_c:
                    extracted.append(end_c)
                self.FlowLine = LineString(extracted)

        # If Z-enabled, report the gradient statistics
        if self.FlowLine.has_z:
            # dh/dl, gradients along path
            self.dhdl = list()
            pt1 = self.FlowLine.coords[0]
            for pt2 in self.FlowLine.coords[1:]:
                self.dhdl.append((pt2[2] - pt1[2]) /
                                 sqrt((pt2[0] - pt1[0]) ** 2 +
                                      (pt2[1] - pt1[1]) ** 2))
                pt1 = pt2
            if len(self.dhdl) == 1:  # line from two points = one slope
                self.sendmsg(' Input line slope (from Z-values) : %12.3g m/m' %
                             self.dhdl[0])
                self.dhdl_mean = self.dhdl_median = self.dhdl[0]
            else:  # more than one slope, show stats
                dhdl_sorted = sorted(self.dhdl)
                n_dhdl = len(self.dhdl)
                if n_dhdl % 2 == 1:
                    self.dhdl_median = dhdl_sorted[n_dhdl // 2]
                else:
                    self.dhdl_median = (dhdl_sorted[n_dhdl // 2 - 1] +
                                        dhdl_sorted[n_dhdl // 2]) / 2.0
                self.dhdl_mean = sum(self.dhdl) / n_dhdl
                self.sendmsg(' Input line slope statistics (from Z-values)')
                self.sendmsg('   number of slopes: %i' % n_dhdl)
                self.sendmsg('          min slope: %12.3g m/m' %
                             dhdl_sorted[0])
                self.sendmsg('       median slope: %12.3g m/m' %
                             self.dhdl_median)
                self.sendmsg('         mean slope: %12.3g m/m' %
                             self.dhdl_mean)
                self.sendmsg('          max slope: %12.3g m/m' %
                             dhdl_sorted[-1])

        # Determine the direction to the stagnation point
        # based on the first two points
        pt1, pt2 = self.FlowLine.coords[0:2]
        self.StagnDir = atan2(pt2[1] - pt1[1], pt2[0] - pt1[0])
        StagnX = self.Well.x + self.x0 * cos(self.StagnDir)
        StagnY = self.Well.y + self.x0 * sin(self.StagnDir)
        if self.Well.has_z:  # Assume downstream point uses hydraulic gradient
            StagnZ = self.Well.z + self.x0 * self.i
            self.StagnPt = Point(StagnX, StagnY, StagnZ)
        else:
            self.StagnPt = Point(StagnX, StagnY)

        self.sendmsg('  Stagnation point: ' + self.StagnPt.wkt)
        self.sendmsg('  Stagnation direction: %.2f degrees' %
                     (self.StagnDir * 180.0 / pi,))

        self.sendmsg(' Ambient flow path line length: %.3f m' %
                     self.FlowLine.length)
        self.sendmsg(' Travel time to line start: %.3f days, %.3f years' %
                     (self.t_days, self.t_years))
        self.sendmsg(' Ambient flow path line Well Known Text (WKT):\n' +
                     self.FlowLine.wkt)

        # Build both left and right sides of the capture zone on
        # the same array, starting with stagnation point in middle
        self.sendmsg('\nBuilding Capture zone polygon...')

        # Calculate y-step size from curve_resolution
        assert self.curve_resolution % 2.0 == 0.0
        dy = self.ymax / 2.0 / self.curve_resolution

        self.sendmsg('Adding stagnation point...')
        self.CZpts = [self.StagnPt.coords[0]]

        # Downstream points, using analytical equation for x
        self.sendmsg('Adding downstream points...')
        DownstreamLine = LineString([self.Well.coords[0],
                                     self.StagnPt.coords[0]])
        self.sendmsg(DownstreamLine.wkt, level=2)
        for yi in range(self.curve_resolution):
            # calculate simple x and y
            yt = dy * (yi + 1)
            xt = self.dslf(yt)
            # Trim off any numerical errors near the well
            if abs(xt) < 0.0001:
                xt = 0.0
            assert xt > self.x0 and xt <= 0.0,\
                '%s not in (%s, %s]' % (xt, self.x0, 0.0)
            # interpolate the point along the capture zone line
            ptx = DownstreamLine.interpolate(-xt)
            #print xt, xt - x0, yt, ptx.wkt
            accumulate_cz_points(ptx, yt, self.StagnDir)

        self.sendmsg(LineString(self.CZpts).wkt, level=2)

        # Calculate upstream points, which is more tricky, since it is like a
        # buffer, except that the width is variable on a non-analytical
        # equation. Make a default buffer width of the second vertex
        xt2 = self.FlowLine.project(Point(self.FlowLine.coords[1]))
        yt2 = self.dslf_i(xt2)

        # Prepare an error message, if needed
        topomsg = dedent("""\
        Cannot build buffer on the %s side, as it is too wide.
        Try changing parameters to reduce the width or straighten the
        capture zone line.""")

        build_side('left')
        build_side('right')

        # Combine capture zone points
        self.sendmsg('Building polygon from points...')
        self.CZpoly = Polygon(shell=LineString(self.CZpts))

        # Fix any bow-tie errors:
        if not self.CZpoly.is_valid:
            self.sendmsg('Repairing geometry with zero buffer...')
            # Apply a zero buffer to fix it
            self.CZpoly = self.CZpoly.buffer(0)
        if not self.CZpoly.is_valid:
            self.sendwarn('Capture zone geometry is not valid')

        # What was the last width?
        self.ylast = self.dslf_i(self.FlowLine.length)

        self.sendmsg('   Capture zone WKT:\n' + self.CZpoly.wkt)
        self.sendmsg('   Number of vertices: %i' %
                     (len(self.CZpoly.boundary.coords) - 1,))
        self.sendmsg('   Capture zone area: %.1f m^2 or %.3f hectares' %
                     (self.CZpoly.area, self.CZpoly.area / 10000.0))
        self.sendmsg('   Capture zone half-width at upstream end: '
                     '%.3f m or %.3f%% of ymax' %
                     (self.ylast, self.ylast / self.ymax * 100.0))

        # Check output
        if self.CZpoly.is_empty:
            raise ValueError('Capture zone Polygon is empty')
        if not self.CZpoly.is_valid:
            self.sendwarn('Capture zone Polygon is not valid')
        if not self.CZpoly.is_simple:
            raise ValueError('Capture zone Polygon is not simple')

        return
