"""GNS Groundwater GeoProcessor Core Tools
"""

__author__ = 'Mike Toews'
__email__ = 'M.Toews@gns.cri.nz'
__license__ = 'Proprietary'
__copyright__ = 'Copyright (c) 2013, ' +\
                'Institute of Geological and Nuclear Sciences Limited ' +\
                '(GNS Science)'

from warnings import warn

# Shapely is used for linear referencing and single-sided buffers
try:
    from shapely.geometry import Point, LineString, Polygon
    from shapely.topology import TopologicalError
except ImportError:
    # Help the user download and install Shapely
    import sys
    pyversion = '.'.join([str(x) for x in sys.version_info[0:2]])
    msg = 'Shapely is required\n\nDownload and install, e.g., ' +\
          'Shapely-x.x.x.' + sys.platform + '-py' + pyversion + '.exe\n' +\
          'from http://www.lfd.uci.edu/~gohlke/pythonlibs/#shapely'
    raise ImportError(msg)

# Set verbosity level for messages between 0 and 2
VERBOSITY = 1


def isnumeric(s):
    """Tests if the object can be cast to a floating point value"""
    try:
        float(s)
    except (ValueError, TypeError):
        return False
    else:
        return True


class GeoProcessor(object):
    """Abstract class for processing geometries

    This class can be inherited to implement the tool in any
    GIS environment that supports Python.
    """
    ## Class members:
    # Curve resolution parameter: more is smoother, less is jagged
    curve_resolution = 16
    days_per_year = 365.25  # some consider this 365.0

    ## Helper functions, which can be overridden by an inherited class
    def sendmsg(self, msg, level=1):
        """Send a message; use level to limit messages by verbosity setting"""
        if level <= VERBOSITY:
            print(msg)

    def sendwarn(self, msg, level=1):
        """Show a warning; use level to limit messages by verbosity setting"""
        if level <= VERBOSITY:
            warn(msg)

    # For errors, raise an exception

    def welcome(self):
        """Show welcome message, where the first line is the
        title of the class doc"""
        self.sendmsg(
            self.__doc__.splitlines()[0] + '\n\n' +
            __copyright__ + '\n' +
            'Written by ' + __author__ + ' <' + __email__ + '>\n')
        return
