#!/usr/bin/env python

"""Setup file for testing and packaging GNS Groundwater Tools

This file does not need to be used for use with ArcGIS, and can be ignored.
ArcGIS is also not required to use these tools.

For Python 2.6 and earlier, install unittest2 from:
http://pypi.python.org/pypi/unittest2

To install unittest2, download the source Python distribution, and
install it from a command prompt with the target version of Python:
> cd unittest2-version
> c:\Python25\python.exe setup.py install

To run tests in this module, use cmd.exe with commands, e.g.:
> c:\Python25\python.exe setup.py test
> c:\Python26\ArcGIS10.0\python.exe setup.py test
> c:\Python27\python.exe setup.py test

For a successful test, should say "OK" at end.

Other commands (among many):
> python setup.py install
> python setup.py sdist
> python setup.py bdist_wininst

Copyright (c) 2013, Institute of Geological and Nuclear Sciences Limited
Written by Mike Toews <M.Toews@gns.cri.nz>

These GIS tools are provided "as is", without warranty of any kind, express or
implied, including but not limited to the warranties of merchantability,
fitness for a particular purpose and noninfringement. In no event shall the
authors or copyright holders be liable for any claim, damages or other
liability, whether in an action of contract, tort or otherwise, arising from,
out of or in connection with the software or the use or other dealings in the
software.
"""

import os
import sys
from distutils.core import setup, Command
if sys.version_info[0:2] < (2, 7):
    try:
        from unittest2 import TextTestRunner, TestLoader
    except ImportError:
        sys.exit('unittest2 required to run tests; see source for details')
else:
    from unittest import TextTestRunner, TestLoader
import shutil
from glob import glob
import shutil

from tests import unit_tests

# Import the module to get version and other info
import ggw

class TestCommand(Command):
    """python setup.py test"""
    user_options = []

    def initialize_options(self):
        self._dir = os.getcwd()

    def finalize_options(self):
        pass

    def run(self):
        """Finds all the tests modules in tests/, and runs them."""
        # Make 'gen' directory, if it doesn't exist
        test_dir = os.path.join(self._dir, 'tests')
        gen_dir = os.path.join(test_dir, 'gen_shapes')
        if not os.path.isdir(gen_dir):
            os.mkdir(gen_dir)
        # Clean up shapefiles in gen dir
        files = glob(os.path.join(gen_dir, '*.shp'))
        files += glob(os.path.join(gen_dir, '*.dbf'))
        files += glob(os.path.join(gen_dir, '*.shx'))
        for file in files:
            try:
                os.unlink(file)
            except:
                pass
        tests = TestLoader().loadTestsFromModule(unit_tests)
        t = TextTestRunner(verbosity=2)
        t.run(tests)


class CleanCommand(Command):
    """python setup.py clean"""
    user_options = []

    def initialize_options(self):
        self._clean_me = []
        for root, dirs, files in os.walk('.'):
            for f in files:
                if f.endswith('.pyc') or f.endswith('.diff'):
                    self._clean_me.append(os.path.join(root, f))

    def finalize_options(self):
        pass

    def run(self):
        for clean_me in self._clean_me:
            try:
                os.unlink(clean_me)
            except:
                pass
        # Hard-coded
        try:
            shutil.rmtree('tests/gen_shapes')
        except:
            pass

setup(
    name          = 'GNS Groundwater Tools',
    version       = ggw.__version__,
    description   = ggw.__doc__,
    author        = ggw.__author__,
    author_email  = ggw.__email__,
    license       = ggw.__license__,
    url           = ggw.__url__,
    # ArcGIS scripts
    scripts       = [
        'ArcGIS_CZCalculatedFixedRadius.py',
        'ArcGIS_CZUniformFlow.py',
    ],
    packages      = ['ggw'],
    cmdclass      = {'test': TestCommand,
                     'clean': CleanCommand},
)
