"""Unit test suite for GNS Groundwater Tools

This module is normally run through ``python setup.py test``,
but it can also be run standalone.
"""

__author__ = 'Mike Toews'
__email__ = 'M.Toews@gns.cri.nz'
__license__ = 'Proprietary'
__copyright__ = 'Copyright (c) 2013, ' +\
                'Institute of Geological and Nuclear Sciences Limited ' +\
                '(GNS Science)'

import os
import sys
import shutil
from pprint import pprint
from glob import glob

if sys.version_info[0:2] < (2, 7):
    import unittest2 as unittest
else:
    import unittest

from math import pi, sqrt

import ggw.core
ggw.core.VERBOSITY = 0

from ggw.CZ import CalculatedFixedRadius, UniformFlow

try:
    from ggw.ArcGIS import ArcGISVersion
    has_ArcGIS = True
except ImportError:
    has_ArcGIS = False

if has_ArcGIS:
    import ArcGIS_CZUniformFlow as CZUniformFlow
    import ArcGIS_CZCalculatedFixedRadius as CZCalculatedFixedRadius
    orig_dir = os.path.join('tests', 'test_shapes')
    gen_dir = os.path.join('tests', 'gen_shapes')

from shapely.wkt import loads as load_wkt
from shapely.topology import TopologicalError
from shapely.geometry import Point


class TestCZCalculatedFixedRadius(unittest.TestCase):

    def test_basic_usage(self):
        Well = load_wkt('POINT(100 200)')
        Q_LPS = 20
        t_years = 2
        n = 0.2
        b = 30
        cz = CalculatedFixedRadius(Well, Q_LPS, t_years, n, b)
        # Verify inputs weren't scrambled
        self.assertEqual(cz.Q_LPS, Q_LPS)
        self.assertEqual(cz.t_years, t_years)
        self.assertEqual(cz.n, n)
        self.assertEqual(cz.b, b)
        # Check derived parameters
        Q = Q_LPS * 24 * 60 * 60 / 1000.0
        self.assertEqual(cz.Q, Q)
        t_days = cz.days_per_year * t_years
        radius = sqrt((Q * t_days) / (pi * n * b))
        self.assertEqual(cz.radius, radius)
        self.assertEqual(cz.Well.coords[:], [(100, 200)])

        # Calculate capture zone geometry
        cz.curve_resolution = 4  # compress the size a bit
        cz.build_cz()
        approxCZpoly = load_wkt('POLYGON ((358.7804225494862 200, '
            '339.0819358080926 100.96901966987444, 282.985391623062 '
            '17.01460837693827, 199.03098033012594 -39.08193580809245, '
            '100 -58.78042254948622, 0.9690196698748395 -39.081935808092766, '
            '-82.98539162306145 17.01460837693773, -139.0819358080923 '
            '100.96901966987367, -158.78042254948622 200, -139.08193580809294 '
            '299.03098033012475, -82.98539162306261 382.9853916230611, '
            '0.9690196698733473 439.08193580809217, 100 458.7804225494862, '
            '199.03098033012444 439.0819358080931, 282.9853916230609 '
            '382.98539162306287, 339.081935808092 299.0309803301271, '
            '358.7804225494862 200))')
        self.assertTrue(cz.CZpoly.almost_equals(approxCZpoly))
        return

    def test_bad_input(self):
        # No defaults
        self.assertRaises(TypeError, CalculatedFixedRadius)
        pt = Point(200.6, 300.2)
        # Wrong number of arguments
        self.assertRaises(TypeError, CalculatedFixedRadius, pt, 20, 5, 0.2)
        self.assertRaises(TypeError, CalculatedFixedRadius, pt, 20, 5, 0.2, 20, 2)
        # But bad values substituted
        self.assertRaises(ValueError, CalculatedFixedRadius, pt, 0, 5, 0.2, 20)
        self.assertRaises(ValueError, CalculatedFixedRadius, pt, 20, 0, 0.2, 20)
        self.assertRaises(ValueError, CalculatedFixedRadius, pt, 20, 5, 0, 20)
        self.assertRaises(ValueError, CalculatedFixedRadius, pt, 20, 5, 0.2, 0)
        self.assertRaises(ValueError, CalculatedFixedRadius, pt, 20, 5, 0, 20)
        self.assertRaises(ValueError, CalculatedFixedRadius, pt, 20, 5, 1, 20)
        # Check geometry input that raise TypeError
        wrong_type = 'POINT(50 60)'
        multi = load_wkt('MULTIPOINT(0 0, 1 2)')
        polygon = load_wkt('POLYGON((50 50, 180 120, 150 -90, 50 50))')
        for geom in [wrong_type, multi, polygon]:
            self.assertRaises(TypeError, CalculatedFixedRadius, geom, 20, 5, 0.2, 20)
        # Check geometry input that raise ValueError
        empty = load_wkt('POINT EMPTY')
        self.assertRaises(ValueError, CalculatedFixedRadius, empty, 20, 5, 0.2, 20)
        return


class TestCZUniformFlow(unittest.TestCase):

    def test_basic_usage(self):
        AB = load_wkt('LINESTRING(0 0, 1000 0)')
        K = 20
        b = 30
        i = 0.006
        Q_LPS = 20
        n = 0.2
        cz = UniformFlow(AB, K, b, i, Q_LPS, n)
        # Verify inputs weren't scrambled
        self.assertTrue(cz.FlowLine.equals(AB))
        self.assertEqual(cz.K, K)
        self.assertEqual(cz.b, b)
        self.assertEqual(cz.i, i)
        self.assertEqual(cz.Q_LPS, Q_LPS)
        self.assertEqual(cz.n, n)
        # Check derived parameters
        T = K * b
        self.assertEqual(cz.T, T)
        Q = Q_LPS * 24 * 60 * 60 / 1000.0
        self.assertEqual(cz.Q, Q)
        x0 = -Q / (2 * pi * K * b * i)
        self.assertEqual(cz.x0, x0)
        ymax = Q / (2 * K * b * i)
        self.assertEqual(cz.ymax, ymax)

        self.assertEqual(AB.length, 1000.0)
        cz.build_cz()
        self.assertEqual(cz.Well.coords[:], [(0.0, 0.0)])
        self.assertEqual(cz.StagnPt.coords[:], [(cz.x0, 0.0)])
        # Calculate capture zone geometry
        cz.curve_resolution = 4  # compress the size a bit
        approxCZpolyFull = load_wkt('POLYGON((1000 223.22214478139335, '
            '238.31953856805916 188.8147631875956, 74.66185178205168 '
            '154.4073815937978, 0 120, -37.27922061357855 90, -60 60, '
            '-72.42640687119285 30, -76.39437268410977 0, '
            '-72.42640687119285 -30, -60 -60, -37.27922061357855 -90, '
            '0 -120, 74.66185178205168 -154.4073815937978, '
            '238.31953856805916 -188.8147631875956, '
            '1000 -223.22214478139335, 1000 223.22214478139335))')
        cz.build_cz()
        self.assertTrue(cz.CZpoly.almost_equals(approxCZpolyFull))
        # Other properties from LineString
        self.assertAlmostEqual(1000.0 / cz.v, 1666.66666666666666)
        self.assertAlmostEqual(cz.t_days, 1329.8358365563927)
        self.assertAlmostEqual(cz.t_years, 3.640892091872396)
        self.assertFalse(hasattr(cz, 'dhdl'))
        self.assertFalse(hasattr(cz, 'dhdl_mean'))
        self.assertFalse(hasattr(cz, 'dhdl_median'))
        self.assertTrue(cz.FlowLine.equals(AB))
        # Use option: time_limit, which modifies the input LineString
        # First case doesn't modify it, since it is too long
        cz.build_cz(5)  # years
        self.assertTrue(cz.CZpoly.almost_equals(approxCZpolyFull))  # previous!
        # Other properties from LineString
        self.assertAlmostEqual(cz.t_days, 1329.8358365563927)
        self.assertAlmostEqual(cz.t_years, 3.640892091872396)
        self.assertTrue(cz.FlowLine.equals(AB))
        # Second case should modify AB to fit 2 years
        approxCZpoly2year = load_wkt('POLYGON((605.5277754283019 '
            '214.0432152764486, 196.06510372380964 182.69547685096572, '
            '65.84180637101126 151.34773842548287, 0 120, '
            '-37.27922061357855 90, -60 60, -72.42640687119285 30, '
            '-76.39437268410977 0, -72.42640687119285 -30, -60 -60, '
            '-37.27922061357855 -90, 0 -120, 65.84180637101126 '
            '-151.34773842548287, 196.06510372380964 -182.69547685096572, '
            '605.5277754283019 -214.0432152764486, 605.5277754283019 '
            '214.0432152764486))')
        cz.build_cz(2)  # years
        self.assertTrue(cz.CZpoly.almost_equals(approxCZpoly2year))
        # Other properties from LineString
        self.assertAlmostEqual(cz.t_days, 730.5, 5)
        self.assertAlmostEqual(cz.t_years, 2.0)
        self.assertFalse(cz.FlowLine.equals(AB))
        self.assertEqual(len(cz.FlowLine.coords[:]), 2)
        self.assertEqual(cz.FlowLine.coords[0], (0.0, 0.0))
        self.assertAlmostEqual(cz.FlowLine.coords[1][0], 605.5277754283019)
        # Retest if original flow path can be re-used, without time_limit
        cz.build_cz()
        self.assertTrue(cz.CZpoly.almost_equals(approxCZpolyFull))
        self.assertAlmostEqual(cz.t_days, 1329.8358365563927)
        return

    def test_Excel_spreadsheet_tool(self):
        AB = load_wkt('LINESTRING(0 0, 1000 0)')  # dummy, not used
        cz = UniformFlow(AB, 20, 30, 0.006, 20, 0.2)
        self.assertEqual(cz.T, 600.0)
        self.assertEqual(cz.Q, 1728.0)
        self.assertAlmostEqual(cz.x0, -76.3943726841098)
        self.assertEqual(cz.ymax, 240.0)
        # Check the computed curve at a few points from Excel spreadsheet
        self.assertAlmostEqual(cz.dslf(0.1), -76.39432905)
        self.assertAlmostEqual(cz.dslf(72), -52.31106202)
        self.assertAlmostEqual(cz.dslf(120), 0.0)
        self.assertAlmostEqual(cz.dslf(235.2) / 1000, 3738.396947 / 1000)
        return

    def test_bad_input(self):
        AB = load_wkt('LINESTRING(0 0, 1000 0)')  # dummy, not used
        # No defaults
        self.assertRaises(TypeError, UniformFlow)
        # Wrong number of arguments
        self.assertRaises(TypeError, UniformFlow, AB, 20, 30, 0.006)
        self.assertRaises(TypeError, UniformFlow, AB, 20, 30, 0.006, 20, 0.2, 2, 5)
        # But bad values substituted
        self.assertRaises(ValueError, UniformFlow, AB, 0, 30, 0.006, 20, 0.2)
        self.assertRaises(ValueError, UniformFlow, AB, 20, 0, 0.006, 20, 0.2)
        self.assertRaises(ValueError, UniformFlow, AB, 20, 30, 0, 20, 0.2)
        self.assertRaises(ValueError, UniformFlow, AB, 20, 30, 0.006, 0, 0.2)
        self.assertRaises(ValueError, UniformFlow, AB, 20, 30, 0.006, 20, 0)
        self.assertRaises(ValueError, UniformFlow, AB, 20, 30, 0.006, 20, 1)
        # Check geometry input that raise TypeError
        wrong_type = 'LINESTRING(50 50, 50 50)'
        polygon = load_wkt('POLYGON((50 50, 180 120, 150 -90, 50 50))')
        for geom in [wrong_type, polygon]:
            self.assertRaises(TypeError, UniformFlow, geom, 20, 30, 0.006, 20, 0.2)
        # Check geometry input that raise ValueError
        nonsimple = load_wkt('LINESTRING(50 20, 360 210, 350 -150, 120 170)')
        ring = load_wkt('LINESTRING(50 50, 180 120, 150 -90, 50 50)')
        invalid = load_wkt('LINESTRING(50 50, 50 50)')
        empty = load_wkt('LINESTRING EMPTY')
        for geom in [nonsimple, ring, invalid, empty]:
            self.assertRaises(ValueError, UniformFlow, geom, 20, 30, 0.006, 20, 0.2)
        # Optional time_limit
        cz = UniformFlow(AB, 20, 30, 0.006, 20, 0.2)
        self.assertRaises(ValueError, cz.build_cz, 'hello')
        self.assertRaises(ValueError, cz.build_cz, 0)
        return

    def test_shapes(self):
        K = 20
        b = 30
        i = 0.006
        Q_LPS = 20
        n = 0.2
        geom = load_wkt('LINESTRING(2789348.0149404076 6344639.285528525, '
            '2787820.8932597586 6345506.968301621, 2785842.5765370997 '
            '6346860.553427651, 2783794.845192593 6348144.723931833, '
            '2782094.1869573244 6348977.699394005, 2780219.992167437 '
            '6349463.601746939, 2779178.7728397218 6349602.4309906345)')
        cz = UniformFlow(geom, K, b, i, Q_LPS, n)
        cz.curve_resolution = 4  # compress the size a bit
        cz.build_cz()
        expected = load_wkt('POLYGON ((2779147.2637296547 6349366.112665131, '
            '2780174.6769592604 6349229.282696332, 2782010.4799504774 '
            '6348753.853157686, 2783679.6652791025 6347936.96124484, '
            '2785713.473645596 6346663.332489463, 2787690.9873476583 '
            '6345317.105814705, 2787707.246065444 6345306.949239627, '
            '2789012.1572764465 6344607.7137768585, 2789199.6568932147 '
            '6344543.371265359, 2789288.7338114944 6344534.9507416375, '
            '2789335.96675654 6344542.618152668, 2789370.5417693947 '
            '6344557.477570625, 2789396.166272562 6344577.422505449, '
            '2789414.4365286855 6344601.54598973, 2789425.8068370186 '
            '6344629.589898893, 2789429.822898308 6344661.812357511, '
            '2789424.8884499096 6344699.120332998, 2789407.2960693208 '
            '6344743.620315412, 2789354.501614568 6344815.788669358, '
            '2789203.4645632827 6344943.776680424, 2787942.85599919 '
            '6345702.20305104, 2785975.6555920774 6347055.053584926, '
            '2785967.7854213305 6347060.21083764, 2783920.8897076077 '
            '6348345.7138341535, 2783899.2005911716 6348357.782870598, '
            '2782198.8111056266 6349191.307030122, 2782153.87926279 '
            '6349207.941143659, 2780279.783972365 6349694.227280232, '
            '2780251.4802109273 6349699.762073116, 2779210.281949789 '
            '6349838.749316138, 2779147.2637296547 6349366.112665131))')
        self.assertTrue(cz.CZpoly.almost_equals(expected))
        self.assertAlmostEqual(cz.t_days, 18446.020220867373)
        self.assertAlmostEqual(cz.t_years, 50.50245098115639)
        cz.build_cz()

    def test_dhdl(self):
        K = 20
        b = 30
        i = 0.006
        Q_LPS = 20
        n = 0.2
        # Two points, 1 dh/dl calculation
        AB = load_wkt('LINESTRING(0 0 0, 1000 0 10)')
        cz = UniformFlow(AB, K, b, i, Q_LPS, n)
        # With a time_limit of 2 years
        cz.build_cz(2)
        self.assertEqual(len(cz.FlowLine.coords[:]), 2)
        c1, c2 = cz.FlowLine.coords[:]
        self.assertEqual(c1, (0.0, 0.0, 0.0))
        self.assertAlmostEqual(c2[0], 605.5277754283019)  # x
        self.assertEqual(c2[1], 0.0)  # y
        self.assertAlmostEqual(c2[2], 6.0552777542830185)  # z
        self.assertEqual(cz.dhdl, [0.01])
        self.assertEqual(cz.dhdl_mean, 0.01)
        self.assertEqual(cz.dhdl_median, 0.01)
        # Now without time_limit
        cz.build_cz()
        self.assertEqual(cz.FlowLine.coords[:],
                         [(0.0, 0.0, 0.0), (1000.0, 0.0, 10.0)])
        self.assertEqual(cz.dhdl, [0.01])
        return

    def test_case1(self):
        # Too short; can't buffer right side
        geom = load_wkt('LINESTRING(2754214.619 6344376.61, '
            '2754266.618 6344379.109, 2754294.418 6344356.609)')
        cz = UniformFlow(geom, 20, 30, 0.006, 20, 0.2)
        self.assertRaises(TopologicalError, cz.build_cz)
        return

    def test_case2(self):
        geom = load_wkt('LINESTRING (2580199.99999994 5960499.99999425, '
            '2580980.3 5958604.5, 2581313.5 5956171.5)')
        cz = UniformFlow(geom, 475, 85, 0.01, 30, .5)
        cz.build_cz()
        return


class TestArcGISCZCalculatedFixedRadius(unittest.TestCase):
    # Test directories with Shapefiles

    def setUp(self):
        if not has_ArcGIS:
            return
        #print(CZCalculatedFixedRadius.gp.productinfo())
        #pprint(CZCalculatedFixedRadius.gp.getinstallinfo())
        # Remove old generated shapefile directory, if there
        self.dir = os.getcwd()
        if os.path.isdir(gen_dir):
            shutil.rmtree(gen_dir)
        os.mkdir(gen_dir)
        CZCalculatedFixedRadius.verbosity = 0

    @unittest.skipIf(not has_ArcGIS, "requires ArcGIS")
    def test_command_line(self):
        # Good values
        FlowLineFeat = os.path.join(orig_dir, 'test_point.shp')
        CZPolyFeat = os.path.join(gen_dir, 'test_shape2_cz.shp')
        Q_LPS = 20
        t_years = 5
        n = 0.2
        b = 30
        # Wrong number of arguments
        wrong_number_argv = ['arg0', 'nonsense']
        self.assertRaises(Exception, CZCalculatedFixedRadius.main,
                          wrong_number_argv)
        # Normal arguments
        argv1 = ['arg0', FlowLineFeat, CZPolyFeat,
                 str(Q_LPS), str(t_years), str(n), str(b)]
        CZCalculatedFixedRadius.main(argv1)
        return


class TestArcGISCZUniformFlow(unittest.TestCase):
    # Test directories with Shapefiles

    def setUp(self):
        if not has_ArcGIS:
            return
        #print(CZUniformFlow.gp.productinfo())
        #pprint(CZUniformFlow.gp.getinstallinfo())
        # Remove old generated shapefile directory, if there
        self.dir = os.getcwd()
        if os.path.isdir(gen_dir):
            shutil.rmtree(gen_dir)
        os.mkdir(gen_dir)
        CZUniformFlow.verbosity = 0

    @unittest.skipIf(not has_ArcGIS, "requires ArcGIS")
    def test_command_line(self):
        # Good values
        FlowLineFeat = os.path.join(orig_dir, 'test_line.shp')
        CZPolyFeat = os.path.join(gen_dir, 'test_shape1_cz.shp')
        K = 20
        b = 30
        i = 0.006
        Q_LPS = 20
        n = 0.2
        time_limit = 100
        # Wrong number of arguments
        wrong_number_argv = ['arg0', 'nonsense']
        self.assertRaises(Exception, CZUniformFlow.main,
                          wrong_number_argv)
        # Normal arguments, without time_limit
        argv1 = ['arg0', FlowLineFeat, CZPolyFeat,
                 str(K), str(b), str(i), str(Q_LPS), str(n)]
        CZUniformFlow.main(argv1)
        # With time_limit as "#" for default
        argv2 = argv1 + ['#']
        CZUniformFlow.main(argv2)
        # With time_limit as a realistic value
        arvg3 = argv1 + ['2']
        CZUniformFlow.main(arvg3)
        return


if __name__ == '__main__':
    """Run unit tests from standalone"""
    print('Running unit tests...')
    print(sys.version)
    unittest.main(verbosity=2)
